/*
 * Code written by Greg Rose, based on the textual description in
 * Schneier's AC2, pp 397-398. Intended to be compatible with RC4,
 * except that the first 256 bytes of output are discarded after key
 * setup.
 *
 * No rights reserved.
 */

#include "arrsyfor.h"

#define S (c->Sbox)
#define SWAP(i,j)   {\
			unsigned char	t;\
			t = S[i];\
			S[i] = S[j];\
			S[j] = t;\
		    }

/*
 * initialise the arrsyfor context, including key setup and initial
 * output disposal (Kocher, I believe).
 */
void
key_arrsyfor(register ctx_arrsyfor *c, unsigned char *key, int n)
{
    register int	i;
    unsigned char	j;
    unsigned char	tbuf[256];

    for (i = 0; i < 256; ++i)
	S[i] = i;
    j = 0;
    for (i = 0; i < 256; ++i) {
	j += S[i] + key[i % n];
	SWAP(i, j);
    }
    c->i = c->j = 0;
    /* burn the first 256 outputs, they aren't as random as one would like */
    arrsyfor(c, tbuf, 256);
}

/*
 * Encrypt buffer by exclusive-or with generated octets. If you really want
 * to see the actual bytes, zero a buffer first.
 */
void
arrsyfor(register ctx_arrsyfor *c, unsigned char *buf, register int n)
{
    register int	i, j;

    i = c->i; j = c->j;
    while (--n >= 0) {
	i = (i + 1) & 0xFF;
	j = (j + S[i]) & 0xFF;
	SWAP(i, j);
	*buf++ ^= S[(S[i]+S[j]) & 0xFF];
    }
    c->i = i; c->j = j;
}

#ifdef TEST
#include <stdio.h>

unsigned char bigbuf[1024*1024];
ctx_arrsyfor	c;

/* test vector courtesy of Adam Back's RC4-in-3-lines-of-perl */

unsigned char expected[] = {
    0xbd, 0xe9, 0x5c, 0xb5, 0x2b, 0x8d, 0xf8, 0xfb,
    0xf2, 0xb7, 0x51, 0xf6, 0x5b, 0xe1, 0xdf, 0x3e,
    0xd7, 0x4b, 0x45, 0x7a, 0xe9, 0x76, 0x4d, 0x26,
    0x2f, 0x43, 0xa4, 0x70, 0x9a, 0x2a, 0xc9, 0x4e,
    0x11, 0x23, 0x89, 0x7b, 0x02, 0x2a, 0x4f, 0x07,
    0x80, 0x98, 0xa1, 0xa0,
};
unsigned char meg_expected[] = {
    0x6b, 0x10, 0xd6, 0x79, 0xc8, 0x87, 0xa1, 0x26,
    0xee, 0x2d, 0x7b, 0xd6, 0xbe, 0x04, 0xbe, 0x0c,
    0x8f, 0x7a, 0xb3, 0xf0, 0xe0, 0xb8, 0xbd, 0xb5,
    0x0f, 0x0c, 0x52, 0x33, 0xae, 0x62, 0xdd, 0x9e,
    0x38, 0x4d, 0x03, 0xdd, 0xaf, 0x56, 0xcd, 0x07,
    0xb1, 0x89, 0x0c, 0x13,
};

void
timekey_arrsyfor(void)
{
    int		i;

    for (i = 0; i < 1000000; ++i)
	key_arrsyfor(&c, (unsigned char *)"test key", 8);
}

void
time_arrsyfor(void)
{
    int		i;

    key_arrsyfor(&c, (unsigned char *)"test key", 8);
    for (i = 0; i < 100; ++i)
	arrsyfor(&c, bigbuf, sizeof bigbuf);
}

int
main(int ac, char **av)
{
    unsigned char	*key;
    int			n, keysz;
    register int	i;
    int			failed = 0;

    if (ac == 2 && strcmp(av[1], "-timekey") == 0) {
	timekey_arrsyfor();
	return 0;
    }
    if (ac == 2 && strcmp(av[1], "-time") == 0) {
	time_arrsyfor();
	return 0;
    }

    if (ac == 2 && strcmp(av[1], "-test") == 0) {
	key_arrsyfor(&c, (unsigned char *)"test key", 8);
	arrsyfor(&c, expected, sizeof expected);
	for (i = 0; i < sizeof expected; ++i) {
	    if (expected[i] != 0) {
		printf("Element %d had discrepency 0x%02X\n", i, expected[i]);
		++failed;
	    }
	}
	arrsyfor(&c, bigbuf, sizeof bigbuf -  sizeof expected);
	arrsyfor(&c, expected, sizeof expected);
	for (i = 0; i < sizeof expected; ++i) {
	    if (expected[i] != meg_expected[i]) {
		printf("Element %d had was 0x%02X, should be 0x%02X\n",
			i + sizeof bigbuf, expected[i], meg_expected[i]);
		++failed;
	    }
	}
	return failed;
    }

    if (ac >= 2)
	key = (unsigned char *)av[1], keysz = strlen(av[1]);
    else 
        key = (unsigned char *)"\0\0\0\0\0\0\0\0", keysz = 8;
    sscanf(ac >= 3 ? av[2] : "10000", "%d", &n);

    key_arrsyfor(&c, key, keysz);
    arrsyfor(&c, bigbuf, n);
    fwrite(bigbuf, n, 1, stdout);
    return 0;
}
#endif
